@@ -3,12 +3,13 @@
  *
  * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
  *
- * Distributed under the Artistic License v2.0; see `doc/COPYING'.
+ * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
 
 #include "config.h"
 #include "pv.h"
 #include <stdio.h>
+#include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
@@ -368,6 +369,9 @@ void display_help(void)
 		{ "-E", "--skip-errors", NULL,
 		 N_("skip read errors in input"),
 		 { 0, 0, 0, 0} },
+		{ "-Z", "--error-skip-block", N_("BYTES"),
+		 N_("skip errors in BYTES blocks at a time"),
+		 { 0, 0, 0, 0} },
 		{ "-S", "--stop-at-size", NULL,
 		 N_("stop after --size bytes have been transferred"),
 		 { 0, 0, 0, 0} },
@@ -377,11 +381,14 @@ void display_help(void)
 		{ "-K", "--direct-io", NULL,
 		 N_("use direct I/O to bypass cache"),
 		 { 0, 0, 0, 0} },
-#ifdef HAVE_IPC
+		{ "-X", "--discard", NULL,
+		 N_("discard input instead of writing to output"),
+		 { 0, 0, 0, 0} },
+#ifdef SA_SIGINFO
 		{ "-R", "--remote", N_("PID"),
 		 N_("update settings of process PID"),
 		 { 0, 0, 0, 0} },
-#endif				/* HAVE_IPC */
+#endif				/* SA_SIGINFO */
 		{ "", NULL, NULL, NULL, { 0, 0, 0, 0} },
 		{ "-P", "--pidfile", N_("FILE"),
 		 N_("save process ID in FILE"),
@@ -424,7 +431,7 @@ void display_help(void)
 	 * be hard to exploit - the message catalogue would have to be
 	 * replaced or forced to load from another location.
 	 */
-	printf(_("Usage: %s [OPTION] [FILE]..."), PROGRAM_NAME);
+	printf(_("Usage: %s [OPTION] [FILE]..."), PACKAGE_NAME);
 	/*@+formatconst@ */
 
 	printf("\n");
@@ -572,14 +579,14 @@ void display_help(void)
 
 	printf("\n");
 
-	bug_report_note = _("Please report any bugs to %s.");
+	bug_report_note = _("Please report any bugs to: %s");
 	if (NULL != bug_report_note) {
 		/*@-formatconst@ */
 		/*
 		 * splint note: see earlier "formatconst" note.
 		 * flawfinder: same reason.
 		 */
-		printf(bug_report_note, BUG_REPORTS_TO);	/* flawfinder: ignore */
+		printf(bug_report_note, PACKAGE_BUGREPORT);	/* flawfinder: ignore */
 		/*@+formatconst@ */
 	}
 
