@@ -4,13 +4,14 @@
  *
  * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
  *
- * Distributed under the Artistic License v2.0; see `doc/COPYING'.
+ * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
 
 #include "config.h"
 #include "options.h"
 #include "pv.h"
 
+/* We do not set this because it breaks "dd" - see below. */
 /* #undef MAKE_STDOUT_NONBLOCKING */
 
 #include <stdio.h>
@@ -18,12 +19,12 @@
 #include <errno.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <termios.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 
 
-int pv_remote_set(opts_t);
+int pv_remote_set(opts_t, pvstate_t);
 void pv_remote_init(void);
 void pv_remote_fini(void);
 
@@ -34,66 +35,144 @@ void pv_remote_fini(void);
  */
 int main(int argc, char **argv)
 {
-	struct termios t, t_save;
-	opts_t opts;
-	pvstate_t state;
-	bool t_saved, t_needs_reset;
+	/*@only@ */ opts_t opts = NULL;
+	/*@only@ */ pvstate_t state = NULL;
 	int retcode = 0;
 
 #ifdef ENABLE_NLS
+	/* Initialise language translation. */
 	(void) setlocale(LC_ALL, "");
 	(void) bindtextdomain(PACKAGE, LOCALEDIR);
 	(void) textdomain(PACKAGE);
 #endif
 
-	opts = opts_parse(argc, argv);
+	/* Parse the command line arguments. */
+	opts = opts_parse(argc >= 0 ? (unsigned int) argc : 0, argv);
 	if (NULL == opts) {
 		debug("%s: %d", "exiting with status", 64);
 		return 64;
 	}
 
+	/* Early exit if necessary, such as with "-h". */
 	if (opts->do_nothing) {
 		debug("%s", "nothing to do - exiting with status 0");
 		opts_free(opts);
 		return 0;
 	}
 
-	/*
-	 * -R specified - send the message, then exit.
-	 */
-	if (opts->remote > 0) {
-		retcode = pv_remote_set(opts);
-		opts_free(opts);
-		return retcode;
-	}
-
 	/*
 	 * Allocate our internal state buffer.
 	 */
 	state = pv_state_alloc(opts->program_name);
 	if (NULL == state) {
+		/*@-mustfreefresh@ */
+		/*
+		 * splint note: the gettext calls made by _() cause memory
+		 * leak warnings, but in this case it's unavoidable, and
+		 * mitigated by the fact we only translate each string once.
+		 */
 		fprintf(stderr, "%s: %s: %s\n", opts->program_name, _("state allocation failed"), strerror(errno));
 		opts_free(opts);
 		debug("%s: %d", "exiting with status", 64);
 		return 64;
+		/*@+mustfreefresh@ */
+	}
+
+	/*
+	 * -R specified - send the message, then exit.
+	 */
+	if (opts->remote > 0) {
+		/* Initialise signal handling. */
+		pv_sig_init(state);
+		/* Send the message. */
+		retcode = pv_remote_set(opts, state);
+		/* Close down the signal handling. */
+		pv_sig_fini(state);
+		/* Free resources. */
+		pv_state_free(state);
+		opts_free(opts);
+		/* Early exit. */
+		return retcode;
 	}
 
 	/*
 	 * Write a PID file if -P was specified.
 	 */
 	if (opts->pidfile != NULL) {
-		FILE *pidfptr;
-		pidfptr = fopen(opts->pidfile, "w");
-		if (NULL == pidfptr) {
-			fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->pidfile, strerror(errno));
+		char *pidfile_tmp_name;
+		size_t pidfile_tmp_bufsize;
+		int pidfile_tmp_fd;
+		FILE *pidfile_tmp_fptr;
+		mode_t prev_umask;
+
+		pidfile_tmp_bufsize = 16 + strlen(opts->pidfile);	/* flawfinder: ignore */
+		/*
+		 * flawfinder rationale: pidfile was supplied as an argument
+		 * so we have to assume it is \0 terminated.
+		 */
+		pidfile_tmp_name = malloc(pidfile_tmp_bufsize);
+		if (NULL == pidfile_tmp_name) {
+			fprintf(stderr, "%s: %s\n", opts->program_name, strerror(errno));
 			pv_state_free(state);
 			opts_free(opts);
 			return 1;
 		}
-		fprintf(pidfptr, "%d\n", getpid());
-		if (0 != fclose(pidfptr)) {
+		memset(pidfile_tmp_name, 0, pidfile_tmp_bufsize);
+		(void) pv_snprintf(pidfile_tmp_name, pidfile_tmp_bufsize, "%s.XXXXXX", opts->pidfile);
+
+		/*@-type@ *//* splint doesn't like mode_t */
+		prev_umask = umask(0000);   /* flawfinder: ignore */
+		(void) umask(prev_umask | 0133);	/* flawfinder: ignore */
+
+		/*@-unrecog@ *//* splint doesn't know mkstemp() */
+		pidfile_tmp_fd = mkstemp(pidfile_tmp_name);	/* flawfinder: ignore */
+		/*@+unrecog@ */
+		if (pidfile_tmp_fd < 0) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, pidfile_tmp_name, strerror(errno));
+			(void) umask(prev_umask);	/* flawfinder: ignore */
+			free(pidfile_tmp_name);
+			pv_state_free(state);
+			opts_free(opts);
+			return 1;
+		}
+
+		(void) umask(prev_umask);   /* flawfinder: ignore */
+
+		/*
+		 * flawfinder rationale (umask, mkstemp) - flawfinder
+		 * recommends setting the most restrictive umask possible
+		 * when calling mkstemp(), so this is what we have done.
+		 *
+		 * We get the original umask and OR it with 0133 to make
+		 * sure new files will be at least chmod 644.  Then we put
+		 * the umask back to what it was, after creating the
+		 * temporary file.
+		 */
+
+		/*@+type@ */
+
+		pidfile_tmp_fptr = fdopen(pidfile_tmp_fd, "w");
+		if (NULL == pidfile_tmp_fptr) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, pidfile_tmp_name, strerror(errno));
+			(void) close(pidfile_tmp_fd);
+			(void) remove(pidfile_tmp_name);
+			free(pidfile_tmp_name);
+			pv_state_free(state);
+			opts_free(opts);
+			return 1;
+		}
+
+		fprintf(pidfile_tmp_fptr, "%d\n", getpid());
+		if (0 != fclose(pidfile_tmp_fptr)) {
 			fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->pidfile, strerror(errno));
 		}
+
+		if (rename(pidfile_tmp_name, opts->pidfile) < 0) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->pidfile, strerror(errno));
+			(void) remove(pidfile_tmp_name);
+		}
+
+		free(pidfile_tmp_name);
 	}
 
 	/*
@@ -101,20 +180,28 @@ int main(int argc, char **argv)
 	 */
 	if (0 == opts->argc) {
 		debug("%s", "no files given - adding fake argument `-'");
-		opts->argv[opts->argc++] = "-";
+		if (!opts_add_file(opts, "-")) {
+			pv_state_free(state);
+			opts_free(opts);
+			return 64;
+		}
 	}
 
 	/*
-	 * Put our list of files into the PV internal state.
+	 * Put our list of input files into the PV internal state.
 	 */
-	pv_state_inputfiles(state, opts->argc, (const char **) (opts->argv));
+	if (NULL != opts->argv) {
+		pv_state_inputfiles(state, opts->argc, (const char **) (opts->argv));
+	}
 
+	/* Total size calculation, in normal transfer mode. */
 	if (0 == opts->watch_pid) {
 		/*
-		 * If no size was given, and we're not in line mode, try to
-		 * calculate the total size.
+		 * If no size was given, try to calculate the total size.
 		 */
-		if ((0 == opts->size) && (false == opts->linemode)) {
+		if (0 == opts->size) {
+			pv_state_linemode_set(state, opts->linemode);
+			pv_state_null_terminated_lines_set(state, opts->null_terminated_lines);
 			opts->size = pv_calc_total_size(state);
 			debug("%s: %llu", "no size given - calculated", opts->size);
 		}
@@ -135,8 +222,8 @@ int main(int argc, char **argv)
 	if ((0 == isatty(STDERR_FILENO))
 	    && (false == opts->force)
 	    && (false == opts->numeric)) {
-		opts->no_op = true;
-		debug("%s", "nothing to display - setting no_op");
+		opts->no_display = true;
+		debug("%s", "nothing to display - setting no_display");
 	}
 
 	/*
@@ -181,9 +268,9 @@ int main(int argc, char **argv)
 	 * Copy parameters from options into main state.
 	 */
 	pv_state_interval_set(state, opts->interval);
-	pv_state_width_set(state, opts->width);
-	pv_state_height_set(state, opts->height);
-	pv_state_no_op_set(state, opts->no_op);
+	pv_state_width_set(state, opts->width, opts->width_set_manually);
+	pv_state_height_set(state, opts->height, opts->height_set_manually);
+	pv_state_no_display_set(state, opts->no_display);
 	pv_state_force_set(state, opts->force);
 	pv_state_cursor_set(state, opts->cursor);
 	pv_state_numeric_set(state, opts->numeric);
@@ -191,11 +278,13 @@ int main(int argc, char **argv)
 	pv_state_delay_start_set(state, opts->delay_start);
 	pv_state_linemode_set(state, opts->linemode);
 	pv_state_bits_set(state, opts->bits);
-	pv_state_null_set(state, opts->null);
+	pv_state_null_terminated_lines_set(state, opts->null_terminated_lines);
 	pv_state_skip_errors_set(state, opts->skip_errors);
+	pv_state_error_skip_block_set(state, opts->error_skip_block);
 	pv_state_stop_at_size_set(state, opts->stop_at_size);
 	pv_state_sync_after_write_set(state, opts->sync_after_write);
 	pv_state_direct_io_set(state, opts->direct_io);
+	pv_state_discard_input_set(state, opts->discard_input);
 	pv_state_rate_limit_set(state, opts->rate_limit);
 	pv_state_target_buffer_size_set(state, opts->buffer_size);
 	pv_state_no_splice_set(state, opts->no_splice);
@@ -220,83 +309,37 @@ int main(int argc, char **argv)
 	fcntl(STDOUT_FILENO, F_SETFL, O_NONBLOCK | fcntl(STDOUT_FILENO, F_GETFL));
 #endif				/* MAKE_STDOUT_NONBLOCKING */
 
-	/*
-	 * Keep track of whether we've saved the terminal attributes and
-	 * whether we need to reset them at the end.
-	 */
-	t_saved = false;
-	t_needs_reset = false;
+	/* Initialise the signal handling. */
+	pv_sig_init(state);
 
-	/*
-	 * Set terminal option TOSTOP so we get signal SIGTTOU if we try to
-	 * write to the terminal while backgrounded.
-	 *
-	 * Also, save the current terminal attributes for later restoration.
-	 */
-	memset(&t, 0, sizeof(t));
-	if (0 != isatty(STDERR_FILENO)) {
-		if (0 == tcgetattr(STDERR_FILENO, &t)) {
-			debug("%s", "saved terminal attributes");
-			t_saved = true;
-		} else {
-			fprintf(stderr, "%s: %s: %s\n", opts->program_name,
-				_("failed to read terminal attributes"), strerror(errno));
-		}
-	}
-	t_save = t;
-	if (t_saved && pv_in_foreground()) {
-		t.c_lflag |= TOSTOP;
-		(void) tcsetattr(STDERR_FILENO, TCSANOW, &t);
-		t_needs_reset = true;
-		debug("%s", "set terminal TOSTOP attribute");
-	}
-
-	if (0 != opts->watch_pid) {
-		if (0 <= opts->watch_fd) {
-			pv_sig_init(state);
-			retcode = pv_watchfd_loop(state);
-			if (t_needs_reset && pv_in_foreground()) {
-				(void) tcsetattr(STDERR_FILENO, TCSANOW, &t_save);
-			}
-			if (opts->pidfile != NULL) {
-				if (0 != remove(opts->pidfile)) {
-					fprintf(stderr, "%s: %s: %s\n",
-						opts->program_name, opts->pidfile, strerror(errno));
-				}
-			}
-			pv_sig_fini(state);
-		} else {
-			pv_sig_init(state);
-			retcode = pv_watchpid_loop(state);
-			if (t_needs_reset && pv_in_foreground()) {
-				(void) tcsetattr(STDERR_FILENO, TCSANOW, &t_save);
-			}
-			if (opts->pidfile != NULL) {
-				if (0 != remove(opts->pidfile)) {
-					fprintf(stderr, "%s: %s: %s\n",
-						opts->program_name, opts->pidfile, strerror(errno));
-				}
-			}
-			pv_sig_fini(state);
-		}
-	} else {
-		pv_sig_init(state);
+	/* Run the appropriate main loop. */
+	if (0 == opts->watch_pid) {
+		/* Normal "transfer data" mode. */
 		pv_remote_init();
 		retcode = pv_main_loop(state);
 		pv_remote_fini();
-		if (t_needs_reset && pv_in_foreground()) {
-			(void) tcsetattr(STDERR_FILENO, TCSANOW, &t_save);
-		}
-		if (opts->pidfile != NULL) {
-			if (0 != remove(opts->pidfile)) {
-				fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->pidfile, strerror(errno));
-			}
-		}
-		pv_sig_fini(state);
+	} else if (0 != opts->watch_pid && -1 == opts->watch_fd) {
+		/* "Watch all file descriptors of another process" mode. */
+		retcode = pv_watchpid_loop(state);
+	} else if (0 != opts->watch_pid && -1 != opts->watch_fd) {
+		/* "Watch a specific file descriptor of another process" mode. */
+		retcode = pv_watchfd_loop(state);
 	}
 
+	/* Clear up the PID file, if one was written. */
+	if (opts->pidfile != NULL) {
+		if (0 != remove(opts->pidfile)) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->pidfile, strerror(errno));
+		}
+	}
+
+	/* Close down the signal handling. */
+	pv_sig_fini(state);
+
+	/* Free the internal PV state. */
 	pv_state_free(state);
 
+	/* Free the data from parsing the command-line arguments. */
 	opts_free(opts);
 
 	debug("%s: %d", "exiting with status", retcode);
