@@ -12,7 +12,7 @@
  *
  * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
  *
- * Distributed under the Artistic License v2.0; see `doc/COPYING'.
+ * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
 
 #include "config.h"
@@ -35,10 +35,12 @@
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
+#endif				/* HAVE_IPC */
+
+/* for basename() */
 #ifdef HAVE_LIBGEN_H
 #include <libgen.h>
 #endif
-#endif				/* HAVE_IPC */
 
 
 /*
@@ -77,34 +79,37 @@ static void pv_crs_open_lockfile(pvstate_t state, int fd)
 	char *tmpdir;
 	int openflags;
 
-	state->crs_lock_fd = -1;
+	state->cursor.lock_fd = -1;
 
 	ttydev = ttyname(fd);
 	if (!ttydev) {
-		if (!state->force) {
+		if (!state->control.force) {
 			pv_error(state, "%s: %s", _("failed to get terminal name"), strerror(errno));
 		}
 		/*
 		 * If we don't know our terminal name, we can neither do IPC
 		 * nor make a lock file, so turn off cursor positioning.
 		 */
-		state->cursor = 0;
+		state->control.cursor = 0;
 		debug("%s", "ttyname failed - cursor positioning disabled");
 		return;
 	}
 
-	tmpdir = (char *) getenv("TMPDIR");
-	if (!tmpdir)
-		tmpdir = (char *) getenv("TMP");
-	if (!tmpdir)
-		tmpdir = "/tmp";
+	/*
+	 * We used to look at the TMPDIR or TMP environment variables to
+	 * override the temporary directory, but this leads to less
+	 * predictable behaviour, and flawfinder points out that relying on
+	 * environment variables in this way is not safe.  So the lock
+	 * directory is hard-coded to "/tmp" now (Sep 2023).
+	 */
+	tmpdir = "/tmp";
 
-	memset(state->crs_lock_file, 0, PV_SIZEOF_CRS_LOCK_FILE);
-	(void) pv_snprintf(state->crs_lock_file,
+	memset(state->cursor.lock_file, 0, PV_SIZEOF_CRS_LOCK_FILE);
+	(void) pv_snprintf(state->cursor.lock_file,
 			   PV_SIZEOF_CRS_LOCK_FILE, "%s/pv-%s-%i.lock", tmpdir, basename(ttydev), (int) geteuid());
 
 	/*
-	 * Pawel Piatek - not everyone has O_NOFOLLOW, e.g. AIX doesn't
+	 * Pawel Piatek - not everyone has O_NOFOLLOW, e.g. AIX doesn't.
 	 */
 #ifdef O_NOFOLLOW
 	openflags = O_RDWR | O_CREAT | O_NOFOLLOW;
@@ -112,10 +117,18 @@ static void pv_crs_open_lockfile(pvstate_t state, int fd)
 	openflags = O_RDWR | O_CREAT;
 #endif
 
-	state->crs_lock_fd = open(state->crs_lock_file, openflags, 0600);
-	if (state->crs_lock_fd < 0) {
-		pv_error(state, "%s: %s: %s", state->crs_lock_file, _("failed to open lock file"), strerror(errno));
-		state->cursor = 0;
+	state->cursor.lock_fd = open(state->cursor.lock_file, openflags, 0600);	/* flawfinder: ignore */
+
+	/*
+	 * flawfinder rationale: we aren't truncating the lock file, we
+	 * don't change its contents, and we are attempting to use
+	 * O_NOFOLLOW where possible to avoid symlink attacks, so this
+	 * open() is as safe as we can make it.
+	 */
+
+	if (state->cursor.lock_fd < 0) {
+		pv_error(state, "%s: %s: %s", state->cursor.lock_file, _("failed to open lock file"), strerror(errno));
+		state->control.cursor = 0;
 		return;
 	}
 }
@@ -131,8 +144,8 @@ static void pv_crs_lock(pvstate_t state, int fd)
 	int lock_fd;
 
 	lock_fd = fd;
-	if (state->crs_lock_fd >= 0)
-		lock_fd = state->crs_lock_fd;
+	if (state->cursor.lock_fd >= 0)
+		lock_fd = state->cursor.lock_fd;
 
 	memset(&lock, 0, sizeof(lock));
 	lock.l_type = (short) F_WRLCK;
@@ -141,10 +154,10 @@ static void pv_crs_lock(pvstate_t state, int fd)
 	lock.l_len = 1;
 	while (fcntl(lock_fd, F_SETLKW, &lock) < 0) {
 		if (errno != EINTR) {
-			if (state->crs_lock_fd == -2) {
+			if (state->cursor.lock_fd == -2) {
 				pv_crs_open_lockfile(state, fd);
-				if (state->crs_lock_fd >= 0) {
-					lock_fd = state->crs_lock_fd;
+				if (state->cursor.lock_fd >= 0) {
+					lock_fd = state->cursor.lock_fd;
 				}
 			} else {
 				pv_error(state, "%s: %s", _("lock attempt failed"), strerror(errno));
@@ -153,8 +166,8 @@ static void pv_crs_lock(pvstate_t state, int fd)
 		}
 	}
 
-	if (state->crs_lock_fd >= 0) {
-		debug("%s: %s", state->crs_lock_file, "terminal lockfile acquired");
+	if (state->cursor.lock_fd >= 0) {
+		debug("%s: %s", state->cursor.lock_file, "terminal lockfile acquired");
 	} else {
 		debug("%s", "terminal lock acquired");
 	}
@@ -171,8 +184,8 @@ static void pv_crs_unlock(pvstate_t state, int fd)
 	int lock_fd;
 
 	lock_fd = fd;
-	if (state->crs_lock_fd >= 0)
-		lock_fd = state->crs_lock_fd;
+	if (state->cursor.lock_fd >= 0)
+		lock_fd = state->cursor.lock_fd;
 
 	memset(&lock, 0, sizeof(lock));
 	lock.l_type = (short) F_UNLCK;
@@ -181,8 +194,8 @@ static void pv_crs_unlock(pvstate_t state, int fd)
 	lock.l_len = 1;
 	(void) fcntl(lock_fd, F_SETLK, &lock);
 
-	if (state->crs_lock_fd >= 0) {
-		debug("%s: %s", state->crs_lock_file, "terminal lockfile released");
+	if (state->cursor.lock_fd >= 0) {
+		debug("%s: %s", state->cursor.lock_file, "terminal lockfile released");
 	} else {
 		debug("%s", "terminal lock released");
 	}
@@ -203,13 +216,13 @@ static void pv_crs_ipccount(pvstate_t state)
 	memset(&buf, 0, sizeof(buf));
 	buf.shm_nattch = 0;
 
-	(void) shmctl(state->crs_shmid, IPC_STAT, &buf);
-	state->crs_pvcount = (int) (buf.shm_nattch);
+	(void) shmctl(state->cursor.shmid, IPC_STAT, &buf);
+	state->cursor.pvcount = (int) (buf.shm_nattch);
 
-	if (state->crs_pvcount > state->crs_pvmax)
-		state->crs_pvmax = state->crs_pvcount;
+	if (state->cursor.pvcount > state->cursor.pvmax)
+		state->cursor.pvmax = state->cursor.pvcount;
 
-	debug("%s: %d", "pvcount", state->crs_pvcount);
+	debug("%s: %d", "pvcount", state->cursor.pvcount);
 }
 #endif				/* HAVE_IPC */
 
@@ -217,12 +230,17 @@ static void pv_crs_ipccount(pvstate_t state)
 /*
  * Get the current cursor Y co-ordinate by sending the ECMA-48 CPR code to
  * the terminal connected to the given file descriptor.
+ *
+ * This should only be called while the terminal is locked, to avoid
+ * confusing other "pv -c" instances about the current terminal attributes,
+ * since this function temporarily changes the terminal attributes and then
+ * puts them back.
  */
 static int pv_crs_get_ypos(int terminalfd)
 {
 	struct termios tty;
 	struct termios old_tty;
-	char cpr[32];
+	char cpr[32];			 /* flawfinder: ignore - bounded, zeroed */
 	int ypos;
 	ssize_t r;
 #ifdef CURSOR_ANSWERBACK_BYTE_BY_BYTE
@@ -247,7 +265,8 @@ static int pv_crs_get_ypos(int terminalfd)
 #ifdef CURSOR_ANSWERBACK_BYTE_BY_BYTE
 	/* Read answerback byte by byte - fails on AIX */
 	for (got = 0, r = 0; got < (int) (sizeof(cpr) - 2); got += r) {
-		r = read(terminalfd, cpr + got, 1);
+		r = read(terminalfd, cpr + got, 1);	/* flawfinder: ignore */
+		/* flawfinder rationale: bounded to buffer size by "for" */
 		if (r <= 0) {
 			debug("got=%d, r=%d: %s", got, r, strerror(errno));
 			break;
@@ -262,7 +281,8 @@ static int pv_crs_get_ypos(int terminalfd)
 
 #else				/* !CURSOR_ANSWERBACK_BYTE_BY_BYTE */
 	/* Read answerback in one big lump - may fail on Solaris */
-	r = read(terminalfd, cpr, sizeof(cpr));
+	r = read(terminalfd, cpr, sizeof(cpr) - 2);	/* flawfinder: ignore */
+	/* flawfinder rationale: bounded to buffer size */
 	if (r <= 0) {
 		debug("r=%d: %s", r, strerror(errno));
 	} else {
@@ -271,9 +291,9 @@ static int pv_crs_get_ypos(int terminalfd)
 		     terminalfd, r, cpr[0], cpr[1], cpr[2], cpr[3], cpr[4], cpr[5]);
 
 	}
-#endif				/* CURSOR_ANSWERBACK_BYTE_BY_BYTE */
+#endif				/* !CURSOR_ANSWERBACK_BYTE_BY_BYTE */
 
-	ypos = (int) pv_getnum_ui(cpr + 2);
+	ypos = (int) pv_getnum_count(cpr + 2);
 
 	if (0 != tcsetattr(terminalfd, TCSANOW | TCSAFLUSH, &old_tty)) {
 		debug("%s: %s", "tcsetattr (2) failed", strerror(errno));
@@ -314,46 +334,51 @@ static int pv_crs_ipcinit(pvstate_t state, char *ttyfile, int terminalfd)
 	}
 
 	pv_crs_lock(state, terminalfd);
-	if (!state->cursor) {
+	if (!state->control.cursor) {
 		debug("%s", "early return - cursor has been disabled");
 		return 1;
 	}
 
-	state->crs_shmid = shmget(key, sizeof(int), 0600 | IPC_CREAT);
-	if (state->crs_shmid < 0) {
+	state->cursor.shmid = shmget(key, sizeof(struct pvcursorstate_s), 0600 | IPC_CREAT);
+	if (state->cursor.shmid < 0) {
 		debug("%s: %s", "shmget failed", strerror(errno));
 		pv_crs_unlock(state, terminalfd);
 		return 1;
 	}
 
-	state->crs_y_top = shmat(state->crs_shmid, NULL, 0);
+	/*@-nullpass@ */
+	/* splint doesn't know shmaddr can be NULL */
+	state->cursor.shared = shmat(state->cursor.shmid, NULL, 0);
+	/*@+nullpass@ */
 
 	pv_crs_ipccount(state);
 
 	/*
 	 * If nobody else is attached to the shared memory segment, we're
 	 * the first, so we need to initialise the shared memory with our
-	 * current Y cursor co-ordinate.
+	 * current Y cursor co-ordinate and with an initial false value for
+	 * the TOSTOP-added flag.
 	 */
-	if (state->crs_pvcount < 2) {
-		state->crs_y_start = pv_crs_get_ypos(terminalfd);
-		*(state->crs_y_top) = state->crs_y_start;
-		state->crs_y_lastread = state->crs_y_start;
+	if (state->cursor.pvcount < 2) {
+		state->cursor.y_start = pv_crs_get_ypos(terminalfd);
+		state->cursor.shared->y_topmost = state->cursor.y_start;
+		state->cursor.shared->tty_tostop_added = false;
+		state->cursor.y_lastread = state->cursor.y_start;
 		debug("%s", "we are the first to attach");
 	}
 
-	state->crs_y_offset = state->crs_pvcount - 1;
-	if (state->crs_y_offset < 0)
-		state->crs_y_offset = 0;
+	state->cursor.y_offset = state->cursor.pvcount - 1;
+	if (state->cursor.y_offset < 0)
+		state->cursor.y_offset = 0;
 
 	/*
 	 * If anyone else had attached to the shared memory segment, we need
 	 * to read the top Y co-ordinate from it.
 	 */
-	if (state->crs_pvcount > 1) {
-		state->crs_y_start = *(state->crs_y_top);
-		state->crs_y_lastread = state->crs_y_start;
-		debug("%s: %d", "not the first to attach - got top y", state->crs_y_start);
+	if (state->cursor.pvcount > 1) {
+		state->cursor.y_start = state->cursor.shared->y_topmost;
+		state->cursor.y_lastread = state->cursor.y_start;
+		debug("%s: %d", "not the first to attach - got top y", state->cursor.y_start);
 	}
 
 	pv_crs_unlock(state, terminalfd);
@@ -369,33 +394,50 @@ static int pv_crs_ipcinit(pvstate_t state, char *ttyfile, int terminalfd)
 void pv_crs_init(pvstate_t state)
 {
 	char *ttyfile;
-	int fd;
+	int terminalfd;
 
-	state->crs_lock_fd = -2;
-	state->crs_lock_file[0] = '\0';
+	state->cursor.lock_fd = -2;
+	state->cursor.lock_file[0] = '\0';
 
-	if (!state->cursor)
+	if (!state->control.cursor)
 		return;
 
 	debug("%s", "init");
 
 	ttyfile = ttyname(STDERR_FILENO);
-	if (!ttyfile) {
+	if (NULL == ttyfile) {
 		debug("%s: %s", "disabling cursor positioning because ttyname failed", strerror(errno));
-		state->cursor = 0;
+		state->control.cursor = false;
 		return;
 	}
 
-	fd = open(ttyfile, O_RDWR);
-	if (fd < 0) {
+	terminalfd = open(ttyfile, O_RDWR); /* flawfinder: ignore */
+
+	/*
+	 * flawfinder rationale: the file we open won't be truncated but
+	 * could be corrupted by writes attempting to get the current Y
+	 * position; but we get the filename from ttyname() and it could be
+	 * a symbolic link, so we can't do much more than trust it.
+	 */
+
+	if (terminalfd < 0) {
 		pv_error(state, "%s: %s: %s", _("failed to open terminal"), ttyfile, strerror(errno));
-		state->cursor = false;
+		state->control.cursor = false;
 		return;
 	}
 #ifdef HAVE_IPC
-	if (pv_crs_ipcinit(state, ttyfile, fd) != 0) {
+	if (pv_crs_ipcinit(state, ttyfile, terminalfd) != 0) {
 		debug("%s", "ipcinit failed, setting noipc flag");
-		state->crs_noipc = true;
+		state->cursor.noipc = true;
+	}
+
+	/*
+	 * If we have already set the terminal TOSTOP attribute, set the
+	 * flag in shared memory to let the other instances know.
+	 */
+	if ((!state->cursor.noipc) && state->signal.pv_tty_tostop_added && (NULL != state->cursor.shared)) {
+		debug("%s", "propagating local pv_tty_tostop_added true value to shared flag");
+		state->cursor.shared->tty_tostop_added = true;
 	}
 
 	/*
@@ -403,29 +445,29 @@ void pv_crs_init(pvstate_t state)
 	 * co-ordinate. If we are using IPC, then the pv_crs_ipcinit()
 	 * function takes care of this in a more multi-process-friendly way.
 	 */
-	if (state->crs_noipc) {
+	if (state->cursor.noipc) {
 #else				/* ! HAVE_IPC */
 	if (1) {
 #endif				/* HAVE_IPC */
 		/*
 		 * Get current cursor position + 1.
 		 */
-		pv_crs_lock(state, fd);
-		state->crs_y_start = pv_crs_get_ypos(fd);
+		pv_crs_lock(state, terminalfd);
+		state->cursor.y_start = pv_crs_get_ypos(terminalfd);
 		/*
 		 * Move down a line while the terminal is locked, so that
 		 * other processes in the pipeline will get a different
 		 * initial ypos.
 		 */
-		if (state->crs_y_start > 0)
+		if (state->cursor.y_start > 0)
 			pv_write_retry(STDERR_FILENO, "\n", 1);
-		pv_crs_unlock(state, fd);
+		pv_crs_unlock(state, terminalfd);
 
-		if (state->crs_y_start < 1)
-			state->cursor = 0;
+		if (state->cursor.y_start < 1)
+			state->control.cursor = 0;
 	}
 
-	(void) close(fd);
+	(void) close(terminalfd);
 }
 
 
@@ -435,9 +477,9 @@ void pv_crs_init(pvstate_t state)
  */
 void pv_crs_needreinit(pvstate_t state)
 {
-	state->crs_needreinit += 2;
-	if (state->crs_needreinit > 3)
-		state->crs_needreinit = 3;
+	state->cursor.needreinit += 2;
+	if (state->cursor.needreinit > 3)
+		state->cursor.needreinit = 3;
 }
 #endif
 
@@ -447,28 +489,29 @@ void pv_crs_needreinit(pvstate_t state)
  * Reinitialise the cursor positioning code (called if we are backgrounded
  * then foregrounded again).
  */
-void pv_crs_reinit(pvstate_t state)
+static void pv_crs_reinit(pvstate_t state)
 {
 	debug("%s", "reinit");
 
 	pv_crs_lock(state, STDERR_FILENO);
 
-	state->crs_needreinit--;
-	if (state->crs_y_offset < 1)
-		state->crs_needreinit = 0;
+	state->cursor.needreinit--;
+	if (state->cursor.y_offset < 1)
+		state->cursor.needreinit = 0;
 
-	if (state->crs_needreinit > 0) {
+	if (state->cursor.needreinit > 0) {
 		pv_crs_unlock(state, STDERR_FILENO);
 		return;
 	}
 
 	debug("%s", "reinit full");
 
-	state->crs_y_start = pv_crs_get_ypos(STDERR_FILENO);
+	state->cursor.y_start = pv_crs_get_ypos(STDERR_FILENO);
 
-	if (state->crs_y_offset < 1)
-		*(state->crs_y_top) = state->crs_y_start;
-	state->crs_y_lastread = state->crs_y_start;
+	if ((state->cursor.y_offset < 1) && (NULL != state->cursor.shared)) {
+		state->cursor.shared->y_topmost = state->cursor.y_start;
+	}
+	state->cursor.y_lastread = state->cursor.y_start;
 
 	pv_crs_unlock(state, STDERR_FILENO);
 }
@@ -476,31 +519,44 @@ void pv_crs_reinit(pvstate_t state)
 
 
 /*
- * Output a single-line update, moving the cursor to the correct position to
- * do so.
+ * Output a single-line update (\0-terminated), using the ECMA-48 CSI "CUP"
+ * sequence to move the cursor to the correct position to do so.
  */
-void pv_crs_update(pvstate_t state, const char *str)
+void pv_crs_update(pvstate_t state, const char *output_line)
 {
-	char pos[32];
+	char cup_cmd[32];		 /* flawfinder: ignore */
+	size_t cup_cmd_length, output_line_length;
 	int y;
 
+	/*
+	 * flawfinder rationale: the "cup_cmd" buffer is always zeroed
+	 * before each use, and is only written to by pv_snprintf() bounded
+	 * by its size; also, that function explictly zero-terminates the
+	 * string it writes.  The write() call is also bounded.
+	 */
+
+	output_line_length = strlen(output_line);	/* flawfinder: ignore */
+	/* flawfinder - output_line is explictly expected to be \0-terminated. */
+
 #ifdef HAVE_IPC
-	if (!state->crs_noipc) {
-		if (state->crs_needreinit > 0)
+	if (!state->cursor.noipc) {
+		if (state->cursor.needreinit > 0)
 			pv_crs_reinit(state);
 
 		pv_crs_ipccount(state);
-		if (state->crs_y_lastread != *(state->crs_y_top)) {
-			state->crs_y_start = *(state->crs_y_top);
-			state->crs_y_lastread = state->crs_y_start;
+		if (NULL != state->cursor.shared) {
+			if (state->cursor.y_lastread != state->cursor.shared->y_topmost) {
+				state->cursor.y_start = state->cursor.shared->y_topmost;
+				state->cursor.y_lastread = state->cursor.y_start;
+			}
 		}
 
-		if (state->crs_needreinit > 0)
+		if (state->cursor.needreinit > 0)
 			return;
 	}
 #endif				/* HAVE_IPC */
 
-	y = state->crs_y_start;
+	y = state->cursor.y_start;
 
 #ifdef HAVE_IPC
 	/*
@@ -509,28 +565,29 @@ void pv_crs_update(pvstate_t state, const char *str)
 	 * scroll the screen (only if we're the first `pv'), and then move
 	 * our initial Y co-ordinate up.
 	 */
-	if (((state->crs_y_start + state->crs_pvmax) > (int) (state->height))
-	    && (!state->crs_noipc)
+	if (((state->cursor.y_start + state->cursor.pvmax) > (int) (state->control.height))
+	    && (!state->cursor.noipc)
 	    ) {
 		int offs;
 
-		offs = ((state->crs_y_start + state->crs_pvmax) - state->height);
+		offs = ((state->cursor.y_start + state->cursor.pvmax) - state->control.height);
 
-		state->crs_y_start -= offs;
-		if (state->crs_y_start < 1)
-			state->crs_y_start = 1;
+		state->cursor.y_start -= offs;
+		if (state->cursor.y_start < 1)
+			state->cursor.y_start = 1;
 
 		debug("%s: %d", "scroll offset", offs);
 
 		/*
 		 * Scroll the screen if we're the first `pv'.
 		 */
-		if (0 == state->crs_y_offset) {
+		if (0 == state->cursor.y_offset) {
 			pv_crs_lock(state, STDERR_FILENO);
 
-			memset(pos, 0, sizeof(pos));
-			(void) pv_snprintf(pos, sizeof(pos), "\033[%u;1H", state->height);
-			pv_write_retry(STDERR_FILENO, pos, strlen(pos));
+			memset(cup_cmd, 0, sizeof(cup_cmd));
+			(void) pv_snprintf(cup_cmd, sizeof(cup_cmd), "\033[%u;1H", state->control.height);
+			cup_cmd_length = strlen(cup_cmd);	/* flawfinder: ignore */
+			pv_write_retry(STDERR_FILENO, cup_cmd, cup_cmd_length);
 			for (; offs > 0; offs--) {
 				pv_write_retry(STDERR_FILENO, "\n", 1);
 			}
@@ -541,24 +598,32 @@ void pv_crs_update(pvstate_t state, const char *str)
 		}
 	}
 
-	if (!state->crs_noipc)
-		y = state->crs_y_start + state->crs_y_offset;
+	if (!state->cursor.noipc)
+		y = state->cursor.y_start + state->cursor.y_offset;
 #endif				/* HAVE_IPC */
 
 	/*
 	 * Keep the Y co-ordinate within sensible bounds, so we can never
-	 * overflow the "pos" buffer.
+	 * overflow the "cup_cmd" buffer.
 	 */
 	if ((y < 1) || (y > 999999))
 		y = 1;
 
-	memset(pos, 0, sizeof(pos));
-	(void) pv_snprintf(pos, sizeof(pos), "\033[%d;1H", y);
+	memset(cup_cmd, 0, sizeof(cup_cmd));
+	(void) pv_snprintf(cup_cmd, sizeof(cup_cmd), "\033[%d;1H", y);
+	cup_cmd_length = strlen(cup_cmd);   /* flawfinder: ignore */
+
+	/*
+	 * flawfinder rationale: "cup_cmd" is only written to by
+	 * pv_snprintf(), which explicitly ensures that the string will be
+	 * \0-terminated, and the buffer is zeroed beforehand so if
+	 * pv_snprintf() fails to run the string will be of zero length.
+	 */
 
 	pv_crs_lock(state, STDERR_FILENO);
 
-	pv_write_retry(STDERR_FILENO, pos, strlen(pos));
-	pv_write_retry(STDERR_FILENO, str, strlen(str));
+	pv_write_retry(STDERR_FILENO, cup_cmd, cup_cmd_length);
+	pv_write_retry(STDERR_FILENO, output_line, output_line_length);
 
 	pv_crs_unlock(state, STDERR_FILENO);
 }
@@ -569,20 +634,22 @@ void pv_crs_update(pvstate_t state, const char *str)
  */
 void pv_crs_fini(pvstate_t state)
 {
-	char pos[32];
+	char cup_cmd[32];		 /* flawfinder: ignore */
 	unsigned int y;
 
+	/* flawfinder - "cup_cmd" is zeroed, and only written by pv_snprintf() */
+
 	debug("%s", "fini");
 
-	y = (unsigned int) (state->crs_y_start);
+	y = (unsigned int) (state->cursor.y_start);
 
 #ifdef HAVE_IPC
-	if ((state->crs_pvmax > 0) && (!state->crs_noipc))
-		y += state->crs_pvmax - 1;
+	if ((state->cursor.pvmax > 0) && (!state->cursor.noipc))
+		y += state->cursor.pvmax - 1;
 #endif				/* HAVE_IPC */
 
-	if (y > state->height)
-		y = state->height;
+	if (y > state->control.height)
+		y = state->control.height;
 
 	/*
 	 * Absolute bounds check.
@@ -590,35 +657,54 @@ void pv_crs_fini(pvstate_t state)
 	if ((y < 1) || (y > 999999))
 		y = 1;
 
-	memset(pos, 0, sizeof(pos));
-	(void) pv_snprintf(pos, sizeof(pos), "\033[%u;1H\n", y);
+	memset(cup_cmd, 0, sizeof(cup_cmd));
+	(void) pv_snprintf(cup_cmd, sizeof(cup_cmd), "\033[%u;1H\n", y);
 
 	pv_crs_lock(state, STDERR_FILENO);
 
-	pv_write_retry(STDERR_FILENO, pos, strlen(pos));
+	pv_write_retry(STDERR_FILENO, cup_cmd, strlen(cup_cmd));	/* flawfinder: ignore */
+	/* flawfinder - pv_snprintf() always \0-terminates (see above). */
 
 #ifdef HAVE_IPC
+	/*
+	 * If any other "pv -c" instances have set the terminal TOSTOP
+	 * attribute, set our local flag so pv_sig_fini() will know about
+	 * it.
+	 */
+	if ((!state->cursor.noipc) && (NULL != state->cursor.shared) && state->cursor.shared->tty_tostop_added) {
+		if (!state->signal.pv_tty_tostop_added) {
+			debug("%s", "propagating shared tty_tostop_added true value to local flag");
+			state->signal.pv_tty_tostop_added = true;
+		}
+	}
+
 	pv_crs_ipccount(state);
-	(void) shmdt((void *) state->crs_y_top);
+	if (NULL != state->cursor.shared) {
+		(void) shmdt(state->cursor.shared);
+	}
+	state->cursor.shared = NULL;
 
 	/*
 	 * If we are the last instance detaching from the shared memory,
 	 * delete it so it's not left lying around.
 	 */
-	if (state->crs_pvcount < 2)
-		(void) shmctl(state->crs_shmid, IPC_RMID, NULL);
+	if (state->cursor.pvcount < 2) {
+		struct shmid_ds shm_buf;
+		memset(&shm_buf, 0, sizeof(shm_buf));
+		(void) shmctl(state->cursor.shmid, IPC_RMID, &shm_buf);
+	}
 
 #endif				/* HAVE_IPC */
 
 	pv_crs_unlock(state, STDERR_FILENO);
 
-	if (state->crs_lock_fd >= 0) {
-		(void) close(state->crs_lock_fd);
+	if (state->cursor.lock_fd >= 0) {
+		(void) close(state->cursor.lock_fd);
 		/*
 		 * We can get away with removing this on exit because all
 		 * the other PVs will be finishing at the same sort of time.
 		 */
-		(void) remove(state->crs_lock_file);
+		(void) remove(state->cursor.lock_file);
 	}
 }
 
